/*
 * I2C simple functions source
 * Version: $Id: i2c.c 147 2009-11-12 23:33:24Z tipok $
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
 
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>

#ifdef HAVE_LINUX_I2C_DEV_H
#include <sys/ioctl.h>
#endif

#ifdef HAVE_SYS_IOCTL_H
#include <linux/i2c-dev.h>
#define DEV_TYPE int
#define i2c_read(a, b, c)	read(a, b, c)
#define i2c_write(a, b, c)	write(a, b, c)
#endif

#ifdef _WIN
#include <usb.h>
#include <windows.h>
#include <winbase.h>
#define DEV_TYPE usb_dev_handle	*
uint8_t addr = -1;
#endif

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "si471x.h"
#include "i2c.h"
#include "logging.h"
#include "telnet.h"



DEV_TYPE deviceId;

#ifdef _WIN

#define USB_CTRL_IN    (USB_TYPE_CLASS | USB_ENDPOINT_IN)
#define USB_CTRL_OUT   (USB_TYPE_CLASS)

/* the vendor and product id was donated by ftdi ... many thanks!*/
#define I2C_TINY_USB_VID  0x0403
#define I2C_TINY_USB_PID  0xc631

#define usleep(t) Sleep((t) / 1000)

#define I2C_M_RD                0x01

/* commands via USB, must e.g. match command ids firmware */
#define CMD_ECHO       0
#define CMD_GET_FUNC   1
#define CMD_SET_DELAY  2
#define CMD_GET_STATUS 3
#define CMD_I2C_IO     4
#define CMD_I2C_BEGIN  1  // flag to I2C_IO
#define CMD_I2C_END    2  // flag to I2C_IO

#define STATUS_IDLE          0
#define STATUS_ADDRESS_ACK   1
#define STATUS_ADDRESS_NAK   2



/* write a set of bytes to the i2c_tiny_usb device */
int i2c_tiny_usb_write(int request, int value, int index)
{
    if(usb_control_msg(deviceId, USB_CTRL_OUT, request, value, index, NULL, 0, 1000) < 0) {
	logwrite(LOG_ERROR, "USB: error: %s", usb_strerror());
	return -1;
    }
    return 1;
}


/* read a set of bytes from the i2c_tiny_usb device */
int i2c_tiny_usb_read(unsigned char cmd, void *data, int len)
{
    int		nBytes;
    
    /* send control request and accept return value */
    nBytes = usb_control_msg(deviceId, USB_CTRL_IN, cmd, 0, 0, data, len, 1000);
    
    if(nBytes < 0) {
	logwrite(LOG_ERROR, "I2C: error: %s", usb_strerror());
	return nBytes;
    }
    return 0;
}


/* get i2c usb interface firmware version */
void i2c_tiny_usb_get_func(void)
{
    unsigned long func;
    
    if(i2c_tiny_usb_read(CMD_GET_FUNC, &func, sizeof(func)) == 0)
	logwrite(LOG_DEBUG,"USB: Functionality = %lx", func);
}


/* set a value in the I2C_USB interface */
void i2c_tiny_usb_set(unsigned char cmd, int value)
{
    if(usb_control_msg(deviceId, USB_TYPE_VENDOR, cmd, value, 0, NULL, 0, 1000) < 0) {
	logwrite(LOG_ERROR, "USB: error: %s", usb_strerror());
    }
}

/* get the current transaction status from the i2c_tiny_usb interface */
int i2c_tiny_usb_get_status(void)
{
    int i;
    unsigned char status;
    
    if((i=i2c_tiny_usb_read(CMD_GET_STATUS, &status, sizeof(status))) < 0) {
	logwrite(LOG_ERROR, "USB: Error reading status");
	return i;
    }
    return status;
}

int i2c_read(usb_dev_handle *deviceId, uint8_t *buf, int amount)
{
    int bytes = 0;
    
    bytes = usb_control_msg(deviceId, USB_CTRL_IN,
			CMD_I2C_IO + 
			CMD_I2C_BEGIN + 
			CMD_I2C_END,
			I2C_M_RD, addr, 
			(char*)buf,
			8, 1000);
    if(bytes < 1) {
        logwrite(LOG_ERROR, "I2C: error: %s", usb_strerror());
        return -1;
    }

    if(i2c_tiny_usb_get_status() != STATUS_ADDRESS_ACK) {
	logwrite(LOG_ERROR, "I2C: read data status failed");
	return -1;
    }

    return bytes;
}

int i2c_write(usb_dev_handle *deviceId, uint8_t *buf, int amount)
{
    int bytes = 0;
    /* write one byte register address to chip */
    bytes = usb_control_msg(deviceId, USB_CTRL_OUT,
			CMD_I2C_IO +
			CMD_I2C_BEGIN +
			CMD_I2C_END,
			0, addr,
			(char *) buf,
			amount, 1000);
    if( bytes < 1) {
	logwrite(LOG_ERROR, "I2C: error: %s", usb_strerror());
	return -1;
    }
    
    if(i2c_tiny_usb_get_status() != STATUS_ADDRESS_ACK) {
	logwrite(LOG_ERROR, "I2C: write command status failed");
	return -1; 
    }
    return bytes;
}

uint8_t device_init(uint8_t i2cbus, uint8_t i2caddr)
{
    deviceId = NULL;
    struct usb_bus      *bus;
    struct usb_device   *dev;
	
    usb_init();
    
    usb_find_busses();
    usb_find_devices();
    uint8_t busnum = 0;
    
    for(bus = usb_get_busses(); bus; bus = bus->next) {
	for(dev = bus->devices; dev; dev = dev->next) {
	    if( (dev->descriptor.idVendor == I2C_TINY_USB_VID) &&
		(dev->descriptor.idProduct == I2C_TINY_USB_PID)) {
		logwrite(LOG_DEBUG, "USB: Found i2c_tiny_usb device on bus %s device %s.",
				bus->dirname, dev->filename);
		
		
		if(busnum != i2cbus){
		    logwrite(LOG_DEBUG, "I2C: Forcing to use other i2c bus busnum=%d, i2cbus=%d", busnum, i2cbus);
		    busnum++;
		    continue;
		}
		
		/* open device */
		if(!(deviceId = usb_open(dev)))
		    logwrite(LOG_ERROR, "USB: Error: Cannot open the device: %s", usb_strerror());
		    break;
	    }
	}
    }
    
    if(!deviceId) {
	logwrite(LOG_ERROR, "USB: Error: Could not find i2c_tiny_usb device");
	return 0;
    }
    
    
    /* do some testing */
    i2c_tiny_usb_get_func();
    
    /* try to set i2c clock to 100kHz (10us), will actually result in ~50kHz
     * since the software generated i2c clock isn't too exact. in fact setting
     * it to 10us doesn't do anything at all since this already is the default */
    i2c_tiny_usb_set(CMD_SET_DELAY, 10);
    
    /* -------- begin of si471x client processing --------- */
    logwrite(LOG_DEBUG, "I2C: Probing for SI471x ... ");
    
    /* try to access si471x at address i2caddr */
    if(usb_control_msg(deviceId, USB_CTRL_IN, CMD_I2C_IO + CMD_I2C_BEGIN + CMD_I2C_END, 0, i2caddr, NULL, 0, 1000) < 0) {
	logwrite(LOG_ERROR, "I2C: error: %s", usb_strerror());
	usb_close(deviceId);
        return 0;
    }
    
    if(i2c_tiny_usb_get_status() == STATUS_ADDRESS_ACK) {
	logwrite(LOG_ERROR, "I2C: success at address 0x%02x", i2caddr);
	addr = i2caddr;
	return 1;
    }
    
    return 0;
}
#else
uint8_t device_init(uint8_t i2cbus, uint8_t i2caddr)
{
    char filename[20];
    deviceId = 0;
    sprintf(filename, "/dev/i2c/%d", i2cbus);
    deviceId = open(filename, O_RDWR);
    if (deviceId < 0 && errno == ENOENT) {
        sprintf(filename, "/dev/i2c-%d", i2cbus);
        deviceId = open(filename, O_RDWR);
    }

    if (deviceId < 0) {
        if (errno == ENOENT) {
            logwrite(LOG_ERROR, "I2C: Error: Could not open file "
                                "`/dev/i2c-%d' or `/dev/i2c/%d': %s\n",
                                i2cbus, i2cbus, strerror(ENOENT));
        } else {
            fprintf(stderr, "I2C: Error: Could not open file "
                            "`%s': %s\n", filename, strerror(errno));
            if (errno == EACCES)
                fprintf(stderr, "I2C: Run as root?\n");
        }
    }

    if (ioctl(deviceId,I2C_SLAVE, i2caddr) < 0) {
        logwrite(LOG_ERROR, "I2C: Can't access to chip: 0x%02x", i2caddr);
        return 0;
    }
    return 1;
}
#endif


uint8_t device_close()
{
	if(deviceId)
#ifdef _WIN
	    usb_close(deviceId);
#else
	    close(deviceId);
#endif
	return 1;
}

/*
 * Print status bits
 */
static uint8_t showStatus(uint8_t status)
{
	logwrite(LOG_XTREME, "STATUS: %s|%s|%s|%s|%s",
					(status & CTS)?"CTS":"cts",
					(status & ERR)?"ERR":"err",
					(status & RDSINT)?"RDSINT":"rdsint",
					(status & ASQINT)?"ASQINT":"asqint",
					(status & STCINT)?"STCINT":"stcint");

	return (status & CTS)?1:0;
}


/*
 * i2c transfer to device (write command and read status and response)
 */
uint8_t transfer(commands_t command, uint8_t *resp, uint8_t amount, ...)
{
	uint8_t buf[9];
	uint8_t i;
	int bytes;
	va_list vl;

#ifdef HAVE_LIBPTHREAD
	if(telnet->enabled) {
	    logwrite(LOG_XTREME, "I2C: Lock");
	    pthread_mutex_lock(&telnet->thread_mutex);

	}
#endif	

	buf[0] = command;
	
	/* variable args handling */
	va_start(vl,amount);
	for (i=1; i<=amount; i++) {
		buf[i] = (uint8_t ) va_arg(vl,int);
	}
	va_end(vl);
	

	logwrite(LOG_XTREME, "I2C: Sending command 0x%02x with %d bytes arguments", command, amount);
	print_bytes(LOG_DUMP, (char *) buf, amount+1);
	
	if(i2c_write(deviceId,buf,amount+1) != amount+1) {
		/* ERROR HANDLING: i2c transaction failed */
		logwrite(LOG_ERROR, "I2C: command write failed");
		exit(1);
	} else {
		logwrite(LOG_XTREME, "I2C: command write OK");
	}
	
	logwrite(LOG_XTREME, "I2C: Trying to read 9 bytes from device:");
	bytes = i2c_read(deviceId,resp,9);
	if (bytes < 1 || bytes > 9){
		logwrite(LOG_ERROR, "I2C: Can't read data from device (received %d bytes)", bytes);
		exit(1);
	}
	
	print_bytes(LOG_DUMP, (char *) resp, bytes);
	showStatus(resp[0]);

#ifdef HAVE_LIBPTHREAD
	if(telnet->enabled) {
	    logwrite(LOG_XTREME, "I2C: UNLock");
	    pthread_mutex_unlock(&telnet->thread_mutex);
	}
#endif	
	
	return resp[0];
}

uint8_t set_property(properties_t addr, uint16_t value)
{
	uint8_t resp[9];
	uint8_t ret;
	
	ret =  transfer(SET_PROPERTY, resp, 5, 0, 
					((addr & 0xFF00) >> 8),
					(addr & 0x00FF),
					((value & 0xFF00) >> 8),
					(value & 0x00FF));
	if ( ret & ERR) {
		logwrite(LOG_ERROR, "I2C: Error setting property 0x%04x to \"0x%04x\". Dying.", addr, value);
		exit(1);
	}
	return ret;
}

uint8_t get_property(properties_t addr, uint16_t *value)
{
	uint8_t resp[9];
	uint8_t ret;
	ret = transfer(GET_PROPERTY, resp, 3, 0, 
					((addr & 0xFF00) >> 8),
					(addr & 0x00FF));
	if ( ret & ERR) {
		logwrite(LOG_ERROR, "I2C: Error getting property 0x%04x. Dying.", addr);
		exit(1);
	}

	*value = (uint16_t ) ((resp[2] & 0x00FF) << 8) | resp[3];

	return ret;
}
