/*
 * Main program to configure si4711
 * Version: $Id: main.c 147 2009-11-12 23:33:24Z tipok $
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <signal.h>

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "si471x.h"
#include "logging.h"
#include "parse_config.h"
#include "telnet.h"

#ifdef _WIN
#define SIGKILL 4
#include <windows.h>  /* for sleep() in dos/mingw */
/* This weird comma expression is because Sleep is void under Windows. */
#include <fcntl.h>
#define sleep(x) Sleep((x)*1000)
#define usleep(x) Sleep((x)/1000)
#define kill(pid, sig) raise (sig)
#endif

si471x_t	*transmitter;
telnet_t	*telnet;

static const char usageStr[] = "Usage: %s -c <firmware file> -v[v...]\n";

static void usage(char *appname) {
	fprintf(stderr, usageStr, appname);
	exit(-1);
}

void CatchInterrupt (int signum) {
	pid_t my_pid;
	
	if (transmitter != NULL){
		if(transmitter->started && transmitter->txoff_onexit)
			si471xShutDown(transmitter);
		si471xClose();
	}
	if(telnet->enabled)
		telnetCloser(telnet);
	
	my_pid = getpid();
	kill(my_pid, SIGKILL);
}



int main(int argc, char **argv)
{
	extern char		*optarg;
	int8_t			ch;
	loglevel=LOG_ERROR;
	transmitter = si471xDefaults();
	telnet = telnetDefaults();
#ifdef HAVE_LIBPTHREAD
	pthread_t	   listenerThread;
	pthread_attr_t listenerAttr;
#endif
	
	while((ch=getopt(argc, argv, "vc:h")) != -1) {
		switch(ch) {
			case 'c':
					parse_config(optarg);
					break;
			case 'v':
					loglevel++;
					break;
			case 'h':
					usage(argv[0]);
					break;
			default:
					logwrite(LOG_ERROR, "Unknown option '%c' 0x%02x", ch, ch);
					usage(argv[0]);
					break;
		}
	}

	if(!si471xOpen(transmitter->bus)){
		logwrite(LOG_INFO, "Can't open chip");
		exit(1);
	}


	if(telnet->enabled) {
#ifdef HAVE_LIBPTHREAD
		pthread_mutex_init(&telnet->thread_mutex, NULL);
		pthread_attr_init(&listenerAttr);
		pthread_attr_setdetachstate(&listenerAttr, PTHREAD_CREATE_DETACHED);
		pthread_create(&listenerThread, &listenerAttr, telnetListener, (void *)telnet);
#else
		logwrite(LOG_INFO, "INFO: Compiled without threads, allowing only 1 telnet connection!");
		telnetListener(telnet);
#endif
	}
	
	si471xInit(transmitter);
	signal(SIGINT, CatchInterrupt);

	
	while (1){
		if(transmitter->started) {
			if(transmitter->rds_enabled && transmitter->rds->send_time)
				si471xRdsTime();
			si471xInterruptHandle(0);
		}
		sleep(5);
	}
	
	
	if(telnet->enabled) {
#ifdef HAVE_LIBPTHREAD
		pthread_attr_destroy(&listenerAttr);
		pthread_mutex_destroy(&telnet->thread_mutex);
#endif
	}
	
	return 1;
}
