/* 

	parse_config.c
	(C) Nicholas J Humfrey <njh@aelius.com> 2006.
	
    Edited: $Id: parse_config.c 147 2009-11-12 23:33:24Z tipok $
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

#include "parse_config.h"
#include "si471x.h"
#include "logging.h"
#include "telnet.h"

/* rds config parsers */
uint16_t parse_rds_audio_type(const char* value, int line_num )
{
	if (!strcmp( "music", value ))			return RDSMS;
	else if (!strcmp( "speech", value ))	return 0x0000;
	else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: rds audio type", line_num);
		exit(-1);
	}

}


uint16_t parse_rds_af(const char* value, int line_num )
{
	uint16_t af = atof(value) * 10;
	if (af == 0)
		return PS_AF_NONE;
	
	if (af > 875 && af < 1080) {
		return (af - 875) | PS_AF1;
	} else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: rds alternative frequency error", line_num);
		exit(-1);
	}
}


uint16_t parse_rds_pi_percentage(const char* value, int line_num ){
	uint16_t percentage = atof(value) * 10;
	if (percentage == 0) return PS_PERCENTAGE_0;
	else if (percentage == 125) return PS_PERCENTAGE_12_5;
	else if (percentage == 250) return PS_PERCENTAGE_25;
	else if (percentage == 500) return PS_PERCENTAGE_50;
	else if (percentage == 750) return PS_PERCENTAGE_75;
	else if (percentage == 875) return PS_PERCENTAGE_87_5;
	else if (percentage == 1000) return PS_PERCENTAGE_100;
	else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: ps percentage value", line_num);
		exit(-1);
	}
	
}



uint16_t parse_rds_pi_code( const char* value, int line_num )
{
	uint16_t result;
	result = strtoul (value, NULL, 0);
	
	if (result == 0) {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: pi code == 0?", line_num);
		/* exit(-1); */
	}
	return result;
}


static void process_statement_rds( char* name, char* value, int line_num )
{

	if (strcmp( "name", name ) == 0) {
		if(transmitter->rds->pi_text != NULL)
			free(transmitter->rds->pi_text);
		transmitter->rds->pi_text = strdup(value);
	} else if (strcmp( "pi-code", name ) == 0) {
		transmitter->rds->pi_code = parse_rds_pi_code(value, line_num);
	} else if (strcmp( "pi-repeats", name ) == 0) {
		transmitter->rds->ps_repeats = atoi( value );
	} else if (strcmp( "send-time", name ) == 0) {
		transmitter->rds->send_time = atoi( value );
	} else if (strcmp( "pi-percentage", name ) == 0) {
		transmitter->rds->ps_mix =  parse_rds_pi_percentage(value, line_num);
	} else if (strcmp( "af", name ) == 0) {
		transmitter->rds->ps_af = parse_rds_af(value, line_num);
	} else if (strcmp( "audio-type", name ) == 0) {
		transmitter->rds->ps_misc =  parse_rds_audio_type(value, line_num);
	} else if (strcmp( "radio-text", name ) == 0) {
		if(transmitter->rds->rt_text != NULL)
			free(transmitter->rds->rt_text);
		transmitter->rds->rt_text = strdup(value);
	} else if (strcmp( "tp", name ) == 0) {
		if (atoi(value) == 1)
			transmitter->rds->ps_misc |= RDSTP;
	} else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid statement in section 'rds'.", line_num);
		exit(-1);
	}
}







/* limiter config parsers */
static void process_statement_limiter( char* name, char* value, int line_num )
{

	if (strcmp( "level-low", name ) == 0) {
		transmitter->limiter->level_low = ((int8_t ) atoi( value )) & 0x00FF;
	} else if (strcmp( "level-high", name ) == 0) {
		transmitter->limiter->level_high = ((int8_t ) atoi( value )) & 0x00FF;;
	} else if (strcmp( "duration-low", name ) == 0) {
		transmitter->limiter->duration_low = atoi( value );
	} else if (strcmp( "duration-high", name ) == 0) {
		transmitter->limiter->duration_high = atoi( value );
	} else if (strcmp( "release-time", name ) == 0) {
		transmitter->limiter->release_time = 512/atof( value );
	} else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid statement in section 'limiter'.", line_num);
		exit(-1);
	}
}



/* compressor config parsers */
acomp_release_t parse_comp_release_time( const char* value, int line_num )
{
	uint16_t reltime = atoi( value );
	if (reltime == 100) return ACOMP_RELEASE_100;
	else if (reltime == 200) return ACOMP_RELEASE_200;
	else if (reltime == 350) return ACOMP_RELEASE_350;
	else if (reltime == 525) return ACOMP_RELEASE_525;
	else if (reltime == 1000) return ACOMP_RELEASE_1000;
	else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid compressor release time", line_num);
		exit(-1);
	}
}


static void process_statement_compressor( char* name, char* value, int line_num )
{

	if (strcmp( "treshold", name ) == 0) {
		transmitter->compressor->treshold = atoi( value );
	} else if (strcmp( "gain", name ) == 0) {
		transmitter->compressor->gain = atoi( value );
	} else if (strcmp( "release-time", name ) == 0) {
		transmitter->compressor->release_time = parse_comp_release_time(value, line_num);
	} else if (strcmp( "attack-time", name ) == 0) {
		transmitter->compressor->attack_time = atof( value )*2 - 1;
	} else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid statement in section 'compressor'.", line_num);
		exit(-1);
	}
}









/* tuning config parsers */
static fmpe_t parse_preemphasis( const char* value, int line_num ) {
	if (!strcmp(value, "50")) return FMPE_50;
	else if (!strcmp(value, "75")) return FMPE_75;
	else if (!strcmp(value, "none")) return FMPE_NONE;
	else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid preemphasis", line_num);
		exit(-1);
	}
}


static void process_statement_tuning( char* name, char* value, int line_num )
{

	if (strcmp( "frequency", name ) == 0) {
		transmitter->radio_out->output_frequency = (uint16_t ) ((float )(atof( value ) * 100.0));
	} else if (strcmp( "power", name ) == 0) {
		transmitter->radio_out->output_power = atoi( value );
	} else if (strcmp( "preemphasis", name ) == 0) {
		transmitter->radio_out->preemphasis = parse_preemphasis(value, line_num);
	} else if (strcmp( "pilot-tone", name ) == 0) {
		transmitter->radio_out->pilot_tone = atoi( value );
	} else if (strcmp( "pilot-devi", name ) == 0) {
		transmitter->radio_out->pilot_deviation = atoi( value );
	} else if (strcmp( "audio-devi", name ) == 0) {
		transmitter->radio_out->audio_deviation = atoi( value );
	} else if (strcmp( "rds-devi", name ) == 0) {
		transmitter->radio_out->rds_deviation = atoi( value );
	} else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid statement in section 'tuning'.", line_num);
		exit(-1);
	}
}





/* audio config parsers */
static asource_t parse_ain( const char* value, int line_num ) {
	if (!strcmp(value, "analog")) return OPMODE_AN;
	else if (!strcmp(value, "digital")) return OPMODE_DI;
	else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid audio input", line_num);
		exit(-1);
	}
}

static uint8_t parse_afmt( const char* value, int line_num ) {
	if (!strcmp(value, "i2s"))			return IMODE_I2S;
	else if (!strcmp(value, "lj"))		return IMODE_LEFT_JUSTIFIED;
	else if (!strcmp(value, "msb1"))	return IMODE_MSB_1DCLK_RE;
	else if (!strcmp(value, "msb2"))	return IMODE_MSB_2DCLK_RE;
	else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid audio format", line_num);
		exit(-1);
	}
}

static mute_t parse_amute( const char* value, int line_num ) {
	if (!strcmp(value, "left")) 		return LIMUTE;
	else if (!strcmp(value, "right"))	return RIMUTE;
	else if (!strcmp(value, "all"))		return ALLMUTE;
	else if (!strcmp(value, "none"))	return NOMUTE;
	else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid audio mute", line_num);
		exit(-1);
	}
}

static uint8_t parse_abps( const char* value, int line_num ) {
	uint8_t bps = atoi(value);
	
	if (bps == 8) 			return ISIZE_8;
	else if (bps == 16) 	return ISIZE_16;
	else if (bps == 20) 	return ISIZE_20;
	else if (bps == 24) 	return ISIZE_24;
	else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid audio bits-per-sample", line_num);
		exit(-1);
	}
}


static void process_statement_audio( char* name, char* value, int line_num )
{

	if (strcmp( "input", name ) == 0) {
		transmitter->audio_input->source = parse_ain(value, line_num);
	} else if (strcmp( "format", name ) == 0) { 
		transmitter->audio_input->format =  parse_afmt(value, line_num);
	} else if (strcmp( "channels", name ) == 0) {
		transmitter->audio_input->channels = atoi( value );
	} else if (strcmp( "samplerate", name ) == 0) { 
		transmitter->audio_input->samplerate = atoi( value );
	} else if (strcmp( "bps", name ) == 0) { 
		transmitter->audio_input->bps = parse_abps(value, line_num);
	} else if (strcmp( "max-level", name ) == 0) { 
		transmitter->audio_input->input_level = atoi( value );
	} else if (strcmp( "mute", name ) == 0) {
		transmitter->audio_input->mute = parse_amute(value, line_num);
	} else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid statement in section 'audio'.", line_num);
		exit(-1);
	}
}



/* chip config parsers */
static void process_statement_chip( char* name, char* value, int line_num )
{

	if (strcmp( "bus", name ) == 0) {
		transmitter->bus = atoi( value );
	} else if (strcmp( "osc-freq", name ) == 0) {
		transmitter->osc_freq = atoi( value );
	} else if (strcmp( "txoff-on-exit", name ) == 0) {
		transmitter->txoff_onexit = atoi( value );
	} else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid statement in section 'chip'.", line_num);
		exit(-1);
	}
}



/* telnet config parsers */
static void process_statement_telnet( char* name, char* value, int line_num )
{

	if (strcmp( "host", name ) == 0) {
		telnet->host = strdup(value);
	} else if (strcmp( "port", name ) == 0) {
		telnet->port = atoi( value );
	} else if (strcmp( "password", name ) == 0) {
		if(strlen(value) > 31) {
                    logwrite(LOG_ERROR, "Too long telnet password on line %d: Max 31 characters allowed.", line_num);
		    exit(-1);
		}
		telnet->password = strdup(value);
	} else if (strcmp( "max-conn", name ) == 0) {
		telnet->max_conn = atoi( value );
	} else {
		logwrite(LOG_ERROR, "Error parsing configuation line %d: invalid statement in section 'telnet'.", line_num);
		exit(-1);
	}
}





/* main parser */
int parse_config( char *filepath )
{

	FILE* file = NULL;
	char line[STR_BUF_SIZE];
	char section[STR_BUF_SIZE];
	char* ptr;
	int i, line_num=0;
	
	// Initialize strings
	line[0] = '\0';
	section[0] = '\0';
	
	
	// Open the input file
	file = fopen( filepath, "r" );
	if (file==NULL) {
		logwrite(LOG_ERROR, "Failed to open config file %s", filepath);
		exit(-1);
	}
	
	transmitter->compressor_enabled = 0;
	transmitter->limiter_enabled	= 0;
	transmitter->rds_enabled		= 0;
	
	// Parse it line by line
	while( !feof( file ) ) {
		line_num++;
		
		fgets( line, STR_BUF_SIZE, file );
		
		// Ignore lines starting with a #
		if (line[0] == '#') continue;
		
		// Remove newline from end
		if (line[strlen(line)-1] == '\n')
			line[strlen(line)-1] = '\0';

		// Ignore empty lines
		if (strlen(line) == 0) continue;
		
		// Is it the start of a section?
		if (line[0]=='[') {
			ptr = &line[1];
			for(i=0; i<strlen(ptr); i++) {
				if (ptr[i] == ']') ptr[i] = '\0';
			}
			
			if (strcmp( ptr, "telnet")==0) {
				strcpy( section, ptr );
				telnet->enabled = 1;
				
			} else if (strcmp( ptr, "chip")==0) {
				strcpy( section, ptr );
				
			} else if (strcmp( ptr, "audio")==0) {
				strcpy( section, ptr );
				
			} else if (strcmp( ptr, "tuning")==0) {
				strcpy( section, ptr );
			
			} else if (strcmp( ptr, "compressor")==0) {
				transmitter->compressor_enabled = 1;
				strcpy( section, ptr );
			
			} else if (strcmp( ptr, "limiter")==0) {
				transmitter->limiter_enabled = 1;
				strcpy( section, ptr );
			
			} else if (strcmp( ptr, "rds")==0) {
				transmitter->rds_enabled = 1;
				strcpy( section, ptr );
			
			} else {
				logwrite(LOG_ERROR, "Error parsing configuation line %d: unknown section '%s'", line_num, ptr);
				exit(-1);
			}
			
			
			
		} else {
			char* name = line;
			char* value = NULL;
			
			// Split up the name and value
			for(i=0; i<strlen(name); i++) {
				if (name[i] == ':') {
					name[i] = '\0';
					value = &name[i+1];
					while (strlen(value) > 0 && (value[0] == ' ' || value[0] == 0x09))
						value++;
					
					if (strlen(value)==0) { value=NULL; }
				}
			}
			
			//fprintf(stderr, "%s: %s=%s.\n", section, name, value);
			
			// Ignore empty values
			if (value==NULL) continue;
			
			if (strcmp( section, "telnet")==0) {	
				process_statement_telnet( name, value, line_num );
			} else if (strcmp( section, "chip")==0) {	
				process_statement_chip( name, value, line_num );
			} else if (strcmp( section, "audio")==0) {	
				process_statement_audio( name, value, line_num );
			} else if (strcmp( section, "tuning")==0) {	
				process_statement_tuning( name, value, line_num );
			} else if (strcmp( section, "compressor")==0) {	
				process_statement_compressor( name, value, line_num );
			} else if (strcmp( section, "limiter")==0) {	
				process_statement_limiter( name, value, line_num );
			} else if (strcmp( section, "rds")==0) {	
				process_statement_rds( name, value, line_num );
			} else {
				logwrite(LOG_ERROR, "Error parsing configuation line %d: missing section.", line_num);
				exit(-1);
			}

		}
		
	}
	

	fclose( file );
	
	return 0;
}

