/*
 * Silabs SI4711 source file (created from AN305 rev 0.6)
 * Version: $Id: si471x.c 149 2010-02-28 13:24:14Z tipok $
 * 
 * High level functions 
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>

#include "si471x.h"
#include "i2c.h"
#include "logging.h"


#ifdef _WIN
#include <windows.h>  /* for sleep() in dos/mingw */
/* This weird comma expression is because Sleep is void under Windows. */
#include <fcntl.h>
#define sleep(x) Sleep((x)*1000)
#define usleep(x) Sleep((x)/1000)
#endif



#define DEFAULT_RDSPS	"Hello :)" \
						"8) World" \
						"From RDS"
#define DEFAULT_RDSRT	"This\nis RadioText\nmessage with newline characters inside"

/*
 * Functions
 * */

int si471xOpen(int8_t bus)
{	
	logwrite(LOG_INFO, "CHIP: trying i2c address 0x%02x", SI4711_I2C_ADDR1);
	if (device_init(bus, SI4711_I2C_ADDR1)) {
		logwrite(LOG_INFO, "CHIP: founded");	
	} else {
		logwrite(LOG_INFO, "CHIP: trying i2c address 0x%02x", SI4711_I2C_ADDR2);
		if (!device_init(bus, SI4711_I2C_ADDR1)) {
			logwrite(LOG_INFO, "CHIP: not found");
			return 0;	
		}
	}
	return 1;
}

int si471xClose()
{
	return device_close();
}


si471x_t *si471xDefaults()
{
	si471x_t		*si471x			= malloc(sizeof(si471x_t));
	ainput_t		*audio_input	= malloc(sizeof(ainput_t));
	radio_t			*radio_out		= malloc(sizeof(radio_t));
	compressor_t	*compressor		= malloc(sizeof(compressor_t));
	limiter_t		*limiter		= malloc(sizeof(limiter_t));
	rds_t			*rds			= malloc(sizeof(rds_t));
	
	
	audio_input->source 		= OPMODE_AN;
	audio_input->channels 		= 2;
	audio_input->samplerate 	= 44100;		/* this 3 params used only */
	audio_input->format 		= IMODE_I2S;	/* digital audio mode */
	audio_input->bps			= ISIZE_16;
	audio_input->input_level	= 636;
	audio_input->mute			= NOMUTE;
	
	radio_out->preemphasis		= FMPE_50;
	radio_out->pilot_tone		= 19000;
	radio_out->audio_deviation	= 66250;
	radio_out->pilot_deviation	= 6750;
	/* used only if rds enabled */
	radio_out->rds_deviation	= 2000;
	radio_out->output_frequency	= 8760; /* 87.6 MHz */
	radio_out->output_power		= 115;
	
	compressor->treshold		= -40;	/* -40 dBFS*/
	compressor->gain			= 15;
	compressor->release_time	= ACOMP_RELEASE_1000;
	compressor->attack_time		= 2;	/*1.5ms    0-9 ->> 0.5ms-5.0ms*/
	
	limiter->level_low			= ((int8_t ) -50) & 0x00FF;	/* –50 dB */
	limiter->level_high			= ((int8_t ) -20) & 0x00FF;	/* –50 dB */
	limiter->duration_low		= 10000;	/* 10000 ms */
	limiter->duration_high		= 5000;		/* 5000 ms */
	limiter->release_time		= 13;		/* (39.38 ms) value=512/39.38 */
	limiter->interrupts			= OVERMODIEN|IALHIEN|IALLIEN;



	rds->interrupts		=	RDSFIFOMT;	/* Interrupt if fifo empty */
	rds->pi_code		=	0x40A7;
	rds->pi_text		=	strdup(DEFAULT_RDSPS);
	rds->ps_mix			=	PS_PERCENTAGE_50;
	rds->ps_misc		=	RDSD0|RDSMS;
	rds->ps_repeats		=	2;
	rds->ps_af			=	PS_AF_NONE;
	rds->fifo_size		=	3+1; /* +1 - needed! (this fifo reserved for sending timestamp)*/
	rds->rt_text		=	strdup(DEFAULT_RDSRT);
	rds->send_time		=	1;

	si471x->started			= 0;
	si471x->chip_id			= 10;	/* default RDS - is turned off,
						 			 * but if detected chip and inner
						 			 * firmware supports it, then it
						 			 * will be switched on automatically */
	si471x->start_mode		= /*CTSIEN|*/FUNC_TX;
	si471x->gp0_interrupts	= 0x00; /* disable GP0 interrupts
					 * sample:GPO_STCIEN|GPO_ERRIEN|GPO_CTSIEN*/
	si471x->osc_freq		= 32768;
	si471x->txoff_onexit		= 1; /* don't turn off transmitter after application exit */

	si471x->audio_input		= audio_input;
	si471x->radio_out		= radio_out;
	
	si471x->compressor_enabled	= 1;
	si471x->compressor		= compressor;

	si471x->limiter_enabled	= 1;
	si471x->limiter			= limiter;

	si471x->rds_enabled			= 1;
	si471x->rds			= rds;

	return si471x;

}

int si471xInit(si471x_t *si471x)
{
	uint8_t 	status;
	uint16_t	buf_flags;
	uint8_t 	resp[9];
	uint16_t	refclk_freq;
	uint16_t	refclk_prescale;
	uint16_t	refclk_remainder;
	
	/* power up */
	logwrite(LOG_DEBUG, "CHIP: powering up device");
	status = transfer(POWER_UP, resp, 2, si471x->start_mode, si471x->audio_input->source);
	if (status & ERR) {
		logwrite(LOG_ERROR, "CHIP:  was turned on before, shuttin it down!");
		if (si471xShutDown(si471x)){
			logwrite(LOG_DEBUG, "CHIP: powering up device again");
			status = transfer(POWER_UP, resp, 2, si471x->start_mode, si471x->audio_input->source);
		}
	}

	if (!(status & CTS)) {
		logwrite(LOG_XTREME, "CHIP: waiting 150ms");
		usleep(150000);
	}

	/* Check Chip/FW/Comp rev GET_REV (command 0x10) */
	logwrite(LOG_DEBUG, "CHIP: getting revision");
	status = transfer(GET_REV, resp, 1, 0x00);
	si471xInterruptHandle(status);
	logwrite(LOG_INFO, "CHIP: si47%d, fw=%c.%c, patch_id=0x%02x.0x%02x, chip_fw=%c.%c, chip_rev=%c",
					resp[1],
					resp[2],resp[3],
					resp[4],resp[5],
					resp[6],resp[7],
					resp[8]);
	if ((resp[1] == 11)||(resp[1] == 13)){
		if (resp[2] < 2){
			logwrite(LOG_ERROR, "CHIP: supports RDS. But firmware (%c.%c) is NOT, please change your chip to more modern one...", resp[2],resp[3]);
		} else {
			si471x->chip_id = resp[1];
			logwrite(LOG_INFO, "CHIP: supports RDS.");
		}
	}

	/* Set GPO */
	set_property(GPO_IEN, si471x->gp0_interrupts);

	/* Set RCLK settings (reference clock and prescaler)*/
	buf_flags = 0x0000;
	if (si471x->osc_freq < 31130 || (si471x->osc_freq > 34406 && si471x->osc_freq < 62260)) {
		logwrite(LOG_ERROR, "CHIP: reference oscillator freq=%d Hz is too low!", si471x->osc_freq);
		exit(1);
	} else if (si471x->osc_freq > 31130 && si471x->osc_freq < 34406) {
		refclk_freq = si471x->osc_freq;
		refclk_prescale = 1;
	} else {
		refclk_freq=31130;
		refclk_remainder = refclk_freq;
		/* search for nominal refclk_freq and refclk_prescale */
		while (refclk_freq <= 34406){
			if (si471x->osc_freq % refclk_freq < refclk_remainder){
				buf_flags = refclk_freq;
				refclk_remainder = si471x->osc_freq % refclk_freq;
				if (refclk_remainder == 0)
					continue;
			}
			refclk_freq++;
		}
		
		if (buf_flags == 0){
			logwrite(LOG_ERROR, "CHIP: can't find refclk and prescaler for oscillator freq=%d Hz!", si471x->osc_freq);
			exit(1);
		}
		
		refclk_freq = buf_flags;
		refclk_prescale = si471x->osc_freq/refclk_freq;
	
	}
	logwrite(LOG_DEBUG, "CHIP: oscillator=%dHz, prescaler=%d, freq=%d Hz", 
					si471x->osc_freq,
					refclk_prescale,
					refclk_freq);
	set_property(REFCLK_FREQ, refclk_freq);
	set_property(REFCLK_PRESCALE, refclk_prescale);
	
	logwrite(LOG_INFO, "CHIP: Init Audio Input...");
	si471xInitAudio(si471x->audio_input);
	
	logwrite(LOG_INFO, "CHIP: Init Tuning Output...");
	si471xInitRadio(si471x->radio_out);
	
	
	buf_flags = 0x0000;
	if ( si471x->compressor_enabled )	buf_flags|= ACEN;
	if ( si471x->limiter_enabled )		buf_flags|= LIMITEN;
	/* enable compressor or limiter */
	logwrite(LOG_DEBUG, "CHIP: Compressor/Limiter flag=0x%04x", buf_flags);
	set_property(TX_ACOMP_ENABLE, buf_flags);

	buf_flags = 0x0000;
	if ( si471x->compressor_enabled ){
		logwrite(LOG_INFO, "CHIP: Init Compressor...");
		si471xInitCompressor(si471x->compressor);
		buf_flags |= OVERMODIEN;
	}

	if ( si471x->limiter_enabled ){
		logwrite(LOG_INFO, "CHIP: Init Limiter...");
		si471xInitLimiter(si471x->limiter);
		buf_flags |= IALHIEN|IALLIEN;
	}
	
	if ( si471x->compressor_enabled || si471x->limiter_enabled){
		/* check input audio signal */
		logwrite(LOG_DEBUG, "CHIP: Compressor/Limiter interrupt flags=0x%04x", buf_flags);
		set_property(TX_ASQ_INTERRUPT_SELECT, buf_flags);
		logwrite(LOG_XTREME, "CHIP: waiting 1s");
		sleep(1);
		si471xInterruptHandle(0);
	}
	
	buf_flags = 0x0000;
	if (si471x->audio_input->channels == 2)	buf_flags |= TX_COMPONENT_LMR|TX_COMPONENT_PILOT;
	
	if ((si471x->chip_id == 11 || si471x->chip_id == 13) && si471x->rds_enabled){
		/* init rds... */
		logwrite(LOG_DEBUG, "TUNING: rds-devi=%d Hz", si471x->radio_out->rds_deviation);
		set_property(TX_RDS_DEVIATION, si471x->radio_out->rds_deviation/10);
		
		si471xInitRDS(si471x->rds, si471x->audio_input->channels);
		
		logwrite(LOG_DEBUG, "TUNING: component_b=0x%04x", si471x->radio_out->rds_deviation);
		buf_flags |= TX_COMPONENT_RDS|TX_COMPONENT_PILOT;
		set_property(TX_COMPONENT_ENABLE, buf_flags);
	} else {
		set_property(TX_COMPONENT_ENABLE, buf_flags);	
	}
	si471x->started = 1;
	return 1;
}

uint8_t	si471xInterruptHandle(uint8_t status_in)
{
	uint8_t		status = 0;
	uint8_t		status_new = 0;
	uint8_t		resp[9];
	uint16_t	resp_freq;
	
	if (status_in & ERR){
		logwrite(LOG_ERROR, "CHIP: answers error (status=0x%02x), check your config!", status_in);
		exit(1);
	}

	if (!(status_in & CTS)){
		status = transfer(GET_INT_STATUS, resp, 0);
		if (!(status & CTS)) {
			logwrite(LOG_ERROR, "CHIP: Can't query device, something wrong?");
			exit(1);
		}
	} else if ((status_in & ASQINT)||(status_in & STCINT)||(status_in & RDSINT)) {
		status = status_in;
	}
	
	/* audio compressor/limiter handler */
	if (status & ASQINT) {
		status_new = transfer(TX_ASQ_STATUS, resp, 1, INTACK);
		if (!(status_new & CTS)) {
			logwrite(LOG_ERROR, "CHIP: getting tune status failed");
			exit(1);
		}
		
		if (resp[1] & OVERMODIEN)
			logwrite(LOG_ERROR, "AUDIO: Overmodulated! (%d dBfs)", (int8_t) resp[4]);
		if (resp[1] & IALHIEN)
			logwrite(LOG_ERROR, "AUDIO: Input Audio Level Threshold Detected High");
		if (resp[1] & IALLIEN)
			logwrite(LOG_ERROR, "AUDIO: Input Audio Level Threshold Detected Low");
	}
	
	/* radio tune handler */
	if (status & STCINT) {
		status_new = transfer(TX_TUNE_STATUS, resp, 1, INTACK);
		if (!(status_new & CTS)) {
			logwrite(LOG_ERROR, "TUNING: getting tune status failed");
			exit(1);
		}
		
		resp_freq = (resp[2] << 8)|resp[3];
		logwrite(LOG_INFO, "TUNING: device tuned: freq=%d.%d MHz, voltage=%d dBuV, ant_cap=%d, rec_nl=%d",
									resp_freq/100, resp_freq%100, resp[5], resp[6], resp[7]);
	}

	if (status & RDSINT) {
		status_new = transfer(TX_RDS_BUFF, resp, 7, INTACK, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00);
		if (resp[1] & RDSPSXMIT)
			logwrite(LOG_DEBUG, "RDS: PS Group has been transmitted.");
		if (resp[1] & RDSCBUFXMIT)
			logwrite(LOG_DEBUG, "RDS: Group has been transmitted from the circular buffer.");
		if (resp[1] & RDSFIFOXMIT)
			logwrite(LOG_DEBUG, "RDS: Group has been transmitted from the FIFO buffer.");
		if (resp[1] & RDSCBUFWRAP)
			logwrite(LOG_DEBUG, "RDS: Group Circular Buffer has wrapped.");
		if (resp[1] & RDSFIFOMT)
			logwrite(LOG_DEBUG, "RDS: Group FIFO Buffer is empty.");
			
		logwrite(LOG_DEBUG, "RDS: circular: %d/%d, fifo: %d/%d", resp[3], resp[2], resp[5], resp[4]);
	}
	return status_new;
}


int	si471xInitCompressor(compressor_t *compressor)
{
	logwrite(LOG_DEBUG, "COMPRESSOR: treshold=%d dBFS, gain=%d dB, release-time_b=0x%04x, attack-time_b=%d",
										compressor->treshold,
										compressor->gain,
										compressor->release_time,
										compressor->attack_time);
	set_property(TX_ACOMP_THRESHOLD, compressor->treshold);
	set_property(TX_ACOMP_GAIN, compressor->gain);
	set_property(TX_ACOMP_RELEASE_TIME, compressor->release_time);
	set_property(TX_ACOMP_ATTACK_TIME, compressor->attack_time);
	return 1;
}

int	si471xInitLimiter(limiter_t *limiter)
{
	logwrite(LOG_DEBUG, "LIMITER: level-low=%d dB, level-high=%d dB, duration-low=%d ms, duration-high=%d ms, release-time=0x%04x",
									(int8_t) limiter->level_low,
									(int8_t) limiter->level_high,
									limiter->duration_low,
									limiter->duration_high,
									limiter->release_time);

	set_property(TX_LIMITER_RELEASE_TIME, limiter->release_time); 
	set_property(TX_ASQ_LOW_LEVEL, limiter->level_low);
	set_property(TX_ASQ_DURATION_LOW, limiter->duration_low);
	set_property(TX_ASQ_HIGH_LEVEL, limiter->level_high);
	set_property(TX_ASQ_DURATION_HIGH, limiter->duration_high);
	return 1;
}

int si471xInitRadio(radio_t *radio_out)
{
	uint8_t status;
	uint8_t resp[9];

	logwrite(LOG_DEBUG, "TUNING: preemphasis_b=0x%02x, pilot-tone=%d Hz, pilot-devi=%d Hz, audio-devi=%d Hz",
							radio_out->preemphasis,
							radio_out->pilot_tone,
							radio_out->pilot_deviation,
							radio_out->audio_deviation);
	set_property(TX_PREEMPHASIS, radio_out->preemphasis);
	set_property(TX_PILOT_FREQUENCY, radio_out->pilot_tone);
	set_property(TX_PILOT_DEVIATION, radio_out->pilot_deviation/10);
	set_property(TX_AUDIO_DEVIATION, radio_out->audio_deviation/10);


	logwrite(LOG_DEBUG, "TUNING: power=%d dBuV, frequency=%d kHz",
								radio_out->output_power,
								radio_out->output_frequency*10);
	
	status = transfer(TX_TUNE_POWER, resp, 4, 0x00, 0x00, radio_out->output_power, RFCAP_AUTO);
	si471xInterruptHandle(status);
	logwrite(LOG_XTREME, "CHIP: waiting 30ms");
	usleep(30000);

	status = transfer(TX_TUNE_FREQ, resp, 3, 0x00, radio_out->output_frequency >> 8, radio_out->output_frequency & 0x00FF);
	si471xInterruptHandle(status);
	logwrite(LOG_XTREME, "CHIP: waiting 120ms");
	usleep(120000);

	status = si471xInterruptHandle(0);
	return status;
}	

int si471xInitAudio(ainput_t *audio_input)
{
	uint16_t	format;
	uint16_t	level;
	uint8_t		atten;
	
	/* setting digital audio parameters */
	if (audio_input->source == OPMODE_DI){
		if (audio_input->channels == 1) {
			format = (audio_input->format << 3)|audio_input->bps|IMONO;
		} else {
			format = (audio_input->format << 3)|audio_input->bps;
		}
		logwrite(LOG_DEBUG, "AUDIO: samplerate=%d, channels=%d, bps_b=0x%02x, format_b=0x%02x",
									audio_input->samplerate,
									audio_input->channels,
									audio_input->bps,
									audio_input->format);
		set_property(DIGITAL_INPUT_SAMPLE_RATE, audio_input->samplerate);
		set_property(DIGITAL_INPUT_FORMAT, format);
	} else {
		if (audio_input->input_level > 416) {
			atten = 3;
			level = 636;
			if (audio_input->input_level > 636)
				logwrite(LOG_ERROR, "AUDIO:  maximum input level is too high (%d mV).",
					audio_input->input_level);
		} else if (audio_input->input_level > 301) {
			atten = 2;
			level = 414;
		} else if (audio_input->input_level > 190) {
			atten = 1;
			level = 301;
		} else {
			atten = 0;
			level = 190;
		}
		logwrite(LOG_DEBUG, "AUDIO: attenuate=%d, max-level=%d mV", atten, audio_input->input_level);
		set_property(TX_LINE_INPUT_LEVEL, (atten << 0x0C)|(level & 0x03FF));
		logwrite(LOG_DEBUG, "AUDIO: mute_b=%d", audio_input->mute);
		set_property(TX_LINE_INPUT_MUTE, audio_input->mute);
	}
	return 1;
}



int si471xInitRDS(rds_t *rds, uint8_t channels)
{
	uint8_t status;
	uint16_t	temp_buff;
	logwrite(LOG_DEBUG, "RDS: interrupts_b=0x%04x", rds->interrupts);
	set_property(TX_RDS_INTERRUPT_SOURCE, rds->interrupts);
	logwrite(LOG_DEBUG, "RDS: pi-code=0x%04x", rds->pi_code);
	set_property(TX_RDS_PI, rds->pi_code);
	logwrite(LOG_DEBUG, "RDS: pi-percentage_b=0x%04x", rds->ps_mix);
	set_property(TX_RDS_PS_MIX, rds->ps_mix);
	
	temp_buff = rds->ps_misc;
	if (channels == 2) temp_buff |= RDSD0;

	logwrite(LOG_DEBUG, "RDS: ps_misc=0x%04x", rds->ps_misc);
	set_property(TX_RDS_PS_MISC, rds->ps_misc);
	logwrite(LOG_DEBUG, "RDS: pi-repeats=%d", rds->ps_repeats);
	set_property(TX_RDS_PS_REPEAT_COUNT, rds->ps_repeats);
	logwrite(LOG_DEBUG, "RDS: af_b=0x%02x", rds->ps_af);
	set_property(TX_RDS_PS_AF, rds->ps_af);
	logwrite(LOG_DEBUG, "RDS: name=\"%s\"", rds->pi_text);
	si471xRdsPS(rds->pi_text);
	logwrite(LOG_DEBUG, "RDS: fifo_size_b=0x%04x", rds->fifo_size);
	set_property(TX_RDS_FIFO_SIZE, rds->fifo_size);
	logwrite(LOG_DEBUG, "RDS: radio-text=\"%s\"", rds->rt_text);
	si471xRdsRadioText(rds->rt_text);
	status = si471xInterruptHandle(0);
	return status;
}


int si471xRdsPS(char *psText)
{
	uint8_t psBlocks, psChars, i, blockPtr;
	char *buf;
	uint8_t resp[9];

	/* counting PS blocks (8 chars)*/
	psBlocks = strlen(psText) / 8;
	if (strlen(psText) % 8)
		psBlocks++;

	psChars = psBlocks*8;
	
	/* is it not much? */
	if (psBlocks > 12 || psBlocks == 0) {
		logwrite(LOG_ERROR, "RDS: PS %d chars - is not good value, max PS length is %d chars", 
												psChars, 12*8);	
		return 0;
	}

	/* Sets the number of program service messages through which to cycle.
	 * Range 1-12 */
	logwrite(LOG_DEBUG, "RDS: PS message count: %d", psBlocks);
	set_property(TX_RDS_PS_MESSAGE_COUNT, psBlocks);

	/* arg 1:
	 * Selects which PS data to load (0–23)
	 * 0 = First 4 characters of PS0.
	 * 1 = Last 4 characters of PS0.
	 * 2 = First 4 characters of PS1.
	 * 3 = Last 4 characters of PS1.
	 * ...
	 * 22 = First 4 characters of PS11.
	 * 23 = Last 4 characters of PS11. */

	/* creating buffer 8'times with whitespaces */
	buf = malloc(psChars);
	memset (buf,' ',psChars);
	
	memcpy(buf, psText, strlen(psText));
	 
	for(i=0;i<psBlocks;i++) {
		blockPtr = i*8;

		logwrite(LOG_DEBUG, "RDS: PS setting chars: '%c%c%c%c'", buf[blockPtr], buf[blockPtr+1], buf[blockPtr+2], buf[blockPtr+3]);
		transfer(TX_RDS_PS, resp, 5, i*2, buf[blockPtr], buf[blockPtr+1], buf[blockPtr+2], buf[blockPtr+3]);
		
		logwrite(LOG_DEBUG, "RDS: PS setting chars: '%c%c%c%c'", buf[blockPtr+4], buf[blockPtr+5], buf[blockPtr+6], buf[blockPtr+7]);
		transfer(TX_RDS_PS, resp, 5, (i*2)+1, buf[blockPtr+4], buf[blockPtr+5], buf[blockPtr+6], buf[blockPtr+7]);
	}
	
	free(buf);
	return 1;
}

int si471xRdsRadioText(char *message)
{
	uint8_t i, avaliableBytes, rtBlocks, rtChars, msgChars, blockPtr;
	uint8_t resp[9];
	char *buf;
	
	msgChars = strlen(message);
	if (msgChars >= 64){
		logwrite(LOG_ERROR, "RDS: RT %d chars - is too long, max RadioText length is %d chars",
							strlen(message), 64-1); 
		exit(1);
	}
	
	/* counting RT blocks (4 chars)*/
	rtBlocks = (msgChars+1)/4;
	if ((msgChars+1) % 4)
		rtBlocks++;

	rtChars = rtBlocks*4;

	/* 1 Block = 4 symbols!
	 * 
	 * have to be 1 block bigger
	 * Default: 0x0000
	 * Units: blocks
	 * Step: 3 blocks
	 * Range: 0, 4, 7, 10..54 */
	 
	transfer(TX_RDS_BUFF, resp, 7, MTBUFF, 0, 0, 0, 0, 0, 0);
	avaliableBytes = resp[2];
	 
	if (rtChars > avaliableBytes) {
		logwrite(LOG_ERROR, "RDS: RT: Avaliable only %d chars, but message creates %d chars, please shorter your radio text message!",
							avaliableBytes, rtChars); 
		exit(1);	
	}
	

	buf = malloc(rtChars);
	memset (buf,0x0d, rtChars);
	memcpy(buf, message, msgChars);

	for(i=0;i<rtBlocks;i++){
		blockPtr = i*4;
		transfer(TX_RDS_BUFF, resp, 7, LDBUFF, 0x20, i, buf[blockPtr], buf[blockPtr+1], buf[blockPtr+2], buf[blockPtr+3]);
		logwrite(LOG_DEBUG, "RDS: circular: %d/%d, fifo: %d/%d", resp[3], resp[2], resp[5], resp[4]);
	}
	
	free(buf);
	return 1;
}

int si471xRdsTime()
{
	uint8_t resp[9];
	uint32_t MJD;
	int y, m, d, k;
	struct tm  *ltm;
	int8_t offset;
	struct timezone tz;
	struct timeval tv;
	gettimeofday( &tv, &tz);
	
	ltm=localtime(&tv.tv_sec);
	offset = tz.tz_minuteswest / 30;
	
	y = ltm->tm_year;
	m = ltm->tm_mon + 1;
	d = ltm->tm_mday;
	k = 0;
	if((m == 1) || (m == 2))
	    k = 1;
	
	MJD = 14956 + d + ((int) ((y - k) * 365.25)) + ((int) ((m + 1 + k * 12) * 30.6001));
	
	/* this is reversed because we have negative value for eastern of GMT */
	if (offset > 0)
	    offset = (abs(offset) & 0x1F) | 0x20;
	else
	    offset = abs(offset) & 0x1F;
	
	/* send timestamp */
	logwrite(LOG_DEBUG, "RDS: local timestamp MJD:%d, %d:%d %s%d/2", MJD, ltm->tm_hour, ltm->tm_min, (offset & 0x20)?"-":"+", offset);
	transfer(TX_RDS_BUFF, resp, 7, FIFO|LDBUFF, 0x40, MJD >> 15, MJD >> 7, MJD << 1|((ltm->tm_hour & 0x1F)>> 4), ((ltm->tm_hour & 0x1F)<< 4)|((ltm->tm_min & 0x3F)>> 2), ((ltm->tm_min & 0x3F)<< 6)|offset);
	return 1;
}


int si471xShutDown(si471x_t *si471x)
{
	uint8_t 	status;
	uint8_t 	resp[9];

	/* carry power down */
	logwrite(LOG_DEBUG, "RADIO: disabling transmission");
	status = transfer(TX_TUNE_POWER, resp, 4, 0x00, 0x00, 0x00, RFCAP_AUTO);
	si471xInterruptHandle(status);
	logwrite(LOG_XTREME, "CHIP: waiting 30ms");
	usleep(30000);

	if (si471x->audio_input->source == OPMODE_DI) {
		logwrite(LOG_DEBUG, "CHIP: turning off Digital Audio");
		set_property(DIGITAL_INPUT_SAMPLE_RATE, 0x0000);
		usleep(30000);
	}

	logwrite(LOG_DEBUG, "CHIP: turning off AFC");
	set_property(REFCLK_FREQ, 0x0000);
	usleep(30000);
	logwrite(LOG_DEBUG, "CHIP: powering down device");
	status = transfer(POWER_DOWN, resp, 0);
	if (status & CTS) {
		logwrite(LOG_XTREME, "CHIP: turning off complete");
	} else if (status & ERR) {
		logwrite(LOG_ERROR, "CHIP: turning off failed!");
		exit(1);
	}
	usleep(150000);
	si471x->started = 0;
	
	return status;
}
