
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifndef _WIN32

#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <netdb.h>
#include <sys/wait.h>

#define sock_write(a,b,c) write(a,b,c)
#define sock_read(a,b,c) read(a,b,c)

#else

#include <winsock2.h>
#define vsnprintf _vsnprintf
#define EINPROGRESS WSAEINPROGRESS
#define ENOTSOCK WSAENOTSOCK
#define EWOULDBLOCK WSAEWOULDBLOCK
#define EALREADY WSAEALREADY
#define socklen_t    int

#define sock_write(a,b,c) send(a,b,c,0);
#define sock_read(a,b,c) recv(a,b,c,0);
#define setsockopt(a,b,c,d,e) setsockopt(a,b,c,(const void*)(d),e)

#ifndef __MINGW32__
#define va_copy(ap1, ap2) memcpy(&ap1, &ap2, sizeof(va_list))
#endif
#endif

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include <signal.h>

#include "telnet.h"
#include "si471x.h"
#include "logging.h"

static const char helpstr[] = "Avaliable commands:\r\n"
" _____________________________________________________\r\n"
"|COMMAND |   Argument   | Description                 |\r\n"
"|________|______________|_____________________________|\r\n"
"|TXFREQ  | 87.5...107.9 | Transmission frequency, Mhz |\r\n"
"|TXPOWER | 0...115      | Transmission power, dBuV    |\r\n"
"|TXOFF   |              | Turn off transmitter        |\r\n"
"|TXON    |              | Turn on transmitter         |\r\n"
"|RDSRT   | <63 chars>   | RDS Radio Text value        |\r\n"
"|RDSPS   | <12*8 chars> | RDS PS Text value           |\r\n"
"|RDSTA   | 0,1          | RDS Traffic Announcement    |\r\n"
"|SHOW    | <command>    | Show value                  |\r\n"
"|HELP    |              | Show this help              |\r\n"
"|EXIT    |              | Disconnect.                 |\r\n"
"|________|______________|_____________________________|\r\n"
"\r\n";

static const char limitstr[] = "Client limit reached!\r\n";

telnet_t *telnetDefaults()
{
	telnet = malloc(sizeof(telnet_t));
	telnet->enabled = 0;
	telnet->sockfd = 0;
	telnet->host = DEFAULT_HOST;
	telnet->port = DEFAULT_PORT;
	telnet->password = NULL;
	telnet->max_conn = DEFAULT_MAX_THREADS;
	telnet->clients = 0;
	telnet->client_pool = NULL;
	return telnet;
}

#ifndef _WIN
static char *strupr (char *a)
{
  char *ret = a;

  while (*a != '\0')
    {
      if (islower (*a))
	*a = toupper (*a);
      ++a;
    }

  return ret;
}
#endif

static void telnetCmd(client_t *client)
{
    char logbuffer[MAXPATHLEN];
    char *param = NULL;
    char *value = NULL;
    int temp, n;
    int exit_code = 0;
    
	do {
		memset(logbuffer,0,MAXPATHLEN);
		temp = 0;

		param = strtok(NULL,"\n");
        if (param == NULL) {

			n = sock_write(client->sockfd,">",1);
			if (n < 0) {
				logwrite(LOG_ERROR, "TELNET: ERROR writing to socket");
				return;
			}

        	n = sock_read(client->sockfd,client->buffer,MAXPATHLEN-1);
        	print_bytes(LOG_DUMP, client->buffer, n);
        	
        	if (n < 0) {
        		logwrite(LOG_ERROR, "TELNET: ERROR reading from socket");
        		return;
        	} else {
        		client->buffer[n] = '\0';
        		client->cmd_ptr = client->buffer;
				param = strtok(client->cmd_ptr,"\n");
        	}
        }

		if(param != NULL) {
        	n=strlen(param);
        	if (n > 0 && param[n-1] == '\x0d') param[n-1] = '\0';

			value = strchr(param,' ');
			while(value && value[0] == ' ') {
				value[0] = '\0';
				value++;
			}

			strupr(param);
		} else {
			continue;
		}
		
		logwrite(LOG_DEBUG, "TELNET: PARAM: '%s', VALUE: '%s'", param, value);
		
		if(strcmp(param, "SHOW") == 0) {
			if (value == NULL || strlen(value) >= 64) {
				snprintf(logbuffer, MAXPATHLEN-1, "bad '%s' value: '%s'\r\n", param, value);
			} else {
				strupr(value);
				if(strcmp(value, "TXFREQ") == 0) {
					snprintf(logbuffer, MAXPATHLEN-1, "%d.%d MHz\r\n",
							transmitter->radio_out->output_frequency/100,
							transmitter->radio_out->output_frequency%100);
				} else if(strcmp(value, "TXPOWER") == 0) {
					snprintf(logbuffer, MAXPATHLEN-1, "%d dBuV\r\n",
							transmitter->radio_out->output_power);
				} else if(strcmp(value, "TXOFF") == 0  || strcmp(value, "TXON") == 0 ) {
					if(strcmp(value, "TXOFF") == 0){
						snprintf(logbuffer, MAXPATHLEN-1, "%d\r\n", !transmitter->started);
					} else {
						snprintf(logbuffer, MAXPATHLEN-1, "%d\r\n", transmitter->started);
					}
				} else if(strcmp(value, "RDSPS") == 0) {
					snprintf(logbuffer, MAXPATHLEN-1, "'%s'\r\n", transmitter->rds->pi_text);
				} else if(strcmp(value, "RDSRT") == 0) {
					snprintf(logbuffer, MAXPATHLEN-1, "'%s'\r\n", transmitter->rds->rt_text);
				} else if(strcmp(value, "RDSTA") == 0) {
					snprintf(logbuffer, MAXPATHLEN-1, "%d\r\n", transmitter->rds->ps_misc & 0x10?1:0);
				} else {
					snprintf(logbuffer, MAXPATHLEN-1, "bad '%s' value: '%s'\r\n", param, value);
				}
			}
		} else if(strcmp(param, "HELP") == 0) {
			n = sock_write(client->sockfd,helpstr,strlen(helpstr));
			if (n < 0) {
				logwrite(LOG_ERROR, "TELNET: ERROR writing to socket");
				return;
			}
		
		} else if(strcmp(param, "TXON") == 0) {
			if(!transmitter->started) {
				temp = si471xInit(transmitter);
				snprintf(logbuffer, MAXPATHLEN-1, "OK %d\r\n", temp);
			} else {
				snprintf(logbuffer, MAXPATHLEN-1, "Transmitter already started!\r\n");
			}
		} else if(strcmp(param, "TXOFF") == 0) {
			if(transmitter->started) {
				temp = si471xShutDown(transmitter);
				snprintf(logbuffer, MAXPATHLEN-1, "OK %d\r\n", temp);
			} else {
				snprintf(logbuffer, MAXPATHLEN-1, "Transmitter already shutted down!\r\n");
			}
			
		} else if(strcmp(param, "TXFREQ") == 0) {
			if(transmitter->started) {
				if (value != NULL)
					temp = ((float) atof(value) * 100);
				
				if (temp > 10800 || temp < 8760) {
					snprintf(logbuffer, MAXPATHLEN-1, "bad '%s' value: %d\r\n", param, temp);
				} else {
					transmitter->radio_out->output_frequency = temp;
					temp = si471xInitRadio(transmitter->radio_out);
					snprintf(logbuffer, MAXPATHLEN-1, "OK %d\r\n", temp);
				}
			} else {
				snprintf(logbuffer, MAXPATHLEN-1, "Transmitter shutted down!\r\n");
			}
		
		} else if (strcmp(param, "TXPOWER") == 0) {
			if(transmitter->started) {
				if (value != NULL)
					temp = atoi(value);
				
				if (temp > 115 || temp < 0) {
					snprintf(logbuffer, MAXPATHLEN-1, "bad '%s' value: %d\r\n", param, temp);
				} else {
					transmitter->radio_out->output_power = temp;
					temp = si471xInitRadio(transmitter->radio_out);
					snprintf(logbuffer, MAXPATHLEN-1, "OK %d\r\n", temp);
				}
			} else {
				snprintf(logbuffer, MAXPATHLEN-1, "Transmitter shutted down!\r\n");
			}
		
		} else if (strcmp(param, "RDSRT") == 0) {
			if(transmitter->started) {
				if (value == NULL || strlen(value) >= 64) {
					snprintf(logbuffer, MAXPATHLEN-1, "bad '%s' value: '%s'\r\n", param, value);
				} else {
					if(transmitter->rds->rt_text != NULL)
						free(transmitter->rds->rt_text);
					transmitter->rds->rt_text = strdup(value);
					temp = si471xRdsRadioText(transmitter->rds->rt_text);
					snprintf(logbuffer, MAXPATHLEN-1, "OK %d\r\n", temp);
				}
			} else {
				snprintf(logbuffer, MAXPATHLEN-1, "Transmitter shutted down!\r\n");
			}
		
		} else if (strcmp(param, "RDSPS") == 0) {
			if(transmitter->started) {
				if (value == NULL || strlen(value) >= 12*8) {
					snprintf(logbuffer, MAXPATHLEN-1, "bad '%s' value: '%s'\r\n", param, value);
				} else {
					if(transmitter->rds->pi_text != NULL)
						free(transmitter->rds->pi_text);
					transmitter->rds->pi_text = strdup(value);
					temp = si471xRdsPS(transmitter->rds->pi_text);
					snprintf(logbuffer, MAXPATHLEN-1, "OK %d\r\n", temp);
				}
			} else {
				snprintf(logbuffer, MAXPATHLEN-1, "Transmitter shutted down!\r\n");
			}

		} else if (strcmp(param, "RDSTA") == 0) {
			if(transmitter->started) {
				if (value != NULL)
					temp = atoi(value);
				
				if (!transmitter->rds_enabled) {
					snprintf(logbuffer, MAXPATHLEN-1, "RDS disabled!\r\n");
				} else if (temp == 0 || temp == 1 ) {
					if (transmitter->chip_id == 11 || transmitter->chip_id == 13) {
						transmitter->rds->ps_misc	&= 0xffef;
						transmitter->rds->ps_misc	|= temp << 4;

						temp = si471xInitRDS(transmitter->rds, transmitter->audio_input->channels);
						snprintf(logbuffer, MAXPATHLEN-1, "OK %d\r\n", temp);
					} else if (!(transmitter->rds->ps_misc & RDSTP)) {
						snprintf(logbuffer, MAXPATHLEN-1, "Can't set Traffic Announcement without Traffic Possible flag, you have to set 'tp' to '1' in 'rds' section at config'\r\n");
					} else {
						snprintf(logbuffer, MAXPATHLEN-1, "Chip id '%c' does not support RDS\r\n", transmitter->chip_id);
					}
				} else {
					snprintf(logbuffer, MAXPATHLEN-1, "bad '%s' value: '%s'\r\n", param, value);
				}
			} else {
				snprintf(logbuffer, MAXPATHLEN-1, "Transmitter shutted down!\r\n");
			}
				
		} else if ( (strcmp(param, "EXIT") == 0) || (strcmp(param, "QUIT") == 0)) {
			return;
		
		} else {
			snprintf(logbuffer, MAXPATHLEN-1, "Unknown command '%s', type 'HELP' to see commands\r\n", param);
			
		}
		
		if(strlen(logbuffer)) {
			n = sock_write(client->sockfd,logbuffer,strlen(logbuffer));
			if (n < 0) {
				logwrite(LOG_ERROR, "TELNET: ERROR writing to socket");
				return;
			}
		}
	} while (!exit_code);
}

void *telnetClient(void *c)
{
    client_t *client = (client_t *) c;
    memset(client->buffer, 0, MAXPATHLEN);
    client->cmd_ptr = client->buffer;
    char *realPasswd = NULL;
    int n;
    
    if (telnet->password != NULL) {
        n = sock_write(client->sockfd,"Password:", 9);
        if (n < 0) {
            logwrite(LOG_ERROR, "TELNET: ERROR writing to socket");
            goto clientFailed;
        }

        n = sock_read(client->sockfd, client->cmd_ptr, MAXPATHLEN);
        if (n < 0) {
            logwrite(LOG_ERROR, "TELNET: ERROR reading from socket");
            goto clientFailed;
        }
        
        print_bytes(LOG_DUMP, client->cmd_ptr, n);
        realPasswd = strtok(client->cmd_ptr,"\n");
        if (realPasswd != NULL) {
        	n=strlen(realPasswd);
        	if (n > 0 && realPasswd[n-1] == '\x0d') realPasswd[n-1] = '\0';
        	if(strcmp(realPasswd, telnet->password) == 0)
            	telnetCmd(client);
        
        } else {
          	sock_write(client->sockfd,"Failed!\r\n", 8);
        }
    } else {
        telnetCmd(client);	
    }
    
clientFailed:    
    close(client->sockfd);
    client->sockfd = -1;
    
    logwrite(LOG_DEBUG, "TELNET: Clients-: %d (clientid: %d)", telnet->clients, client->id);
    telnet->clients--;

#ifdef HAVE_LIBPTHREAD
    pthread_exit(NULL);
#endif
    return NULL;
}


void *telnetListener(void *t)
{
	telnet_t *telnet = (telnet_t *) t;
	struct hostent *loHost;
	struct sockaddr_in serv_addr;
	int i;
#ifdef HAVE_LIBPTHREAD
	int rc;
#endif
#ifdef _WIN
        WSADATA wsadata;
#endif
	
	telnet->client_pool = calloc(telnet->max_conn+1, sizeof(void *));
		
	for(i=0; i <= telnet->max_conn; i++) {
		telnet->client_pool[i] = malloc(sizeof(client_t));
		telnet->client_pool[i]->sockfd = -1;
		telnet->client_pool[i]->socketlen = sizeof(struct sockaddr_in);
		telnet->client_pool[i]->id = i;
#ifdef HAVE_LIBPTHREAD
		pthread_attr_init(&telnet->client_pool[i]->attr);
		pthread_attr_setdetachstate(&telnet->client_pool[i]->attr, PTHREAD_CREATE_DETACHED);
#endif
	}
	
#ifdef _WIN
    if (WSAStartup(MAKEWORD(2,2), &wsadata)) {
	logwrite(LOG_ERROR, "TELNET: unable to initialize winsock subsystem, error %d",WSAGetLastError());
	exit(-1);
    }
#endif

    loHost = gethostbyname(telnet->host);
    if (loHost == NULL) {
		logwrite(LOG_ERROR, "TELNET: bad hostname!");
		exit(1);
	}
	

#ifdef _WIN
    if (WSAStartup(MAKEWORD(1,1), &wsadata) == SOCKET_ERROR) {
	logwrite(LOG_ERROR, "TELNET: Error creating WIN socket.");
        exit(1);
    }
#endif

#ifdef _WIN
    telnet->sockfd = WSASocket(AF_INET, SOCK_STREAM, IPPROTO_TCP, NULL, 0, 0);
    if (telnet->sockfd == INVALID_SOCKET) {
	/*
	 * WSAGetLastError() values are regular BSD error codes
	 * biased by WSABASEERR.
	 * However, strerror() does not know about networking
	 * specific errors, which are values beginning at 38 or so.
	 * Therefore, we choose to leave the biased error code
	 * in errno so that _if_ someone looks up the code somewhere,
	 * then it is at least the number that are usually listed.
	 */
	errno = WSAGetLastError();
    }
#if 0
    /* convert into a file descriptor */
    if ((telnet->sockfd = _open_osfhandle(s, O_RDWR|O_BINARY)) < 0) {
	closesocket(s);
	logwrite(LOG_ERROR, "TELNET: unable to make WIN32 socket file descriptor: %s", strerror(errno));
        exit(1);
    }
#endif
#else
    /* First call to socket() function */
    telnet->sockfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
#endif

    if (telnet->sockfd < 0) {
#ifdef _WIN
	errno = WSAGetLastError();
#endif
        logwrite(LOG_ERROR, "TELNET: ERROR opening socket %s", strerror(errno));
        exit(1);
    }

    /* enable socket reuse */
    i = 1;
    if ( setsockopt(telnet->sockfd, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i) ) <0 )
    {
    	logwrite(LOG_ERROR, "TELNET: ERROR set reuse option for %d socket %s", i, strerror(errno));
    	exit(1);
    }
    
    /* Initialize socket structure */
    memset((char *) &serv_addr, 0, sizeof(struct sockaddr_in));
    serv_addr.sin_family = AF_INET;

    memcpy((char *) &serv_addr.sin_addr.s_addr,
	   (char *)loHost->h_addr, 
                loHost->h_length);
    
    serv_addr.sin_port = htons(telnet->port);
 
    /* Now bind the host address using bind() call.*/
    if (bind( telnet->sockfd, (struct sockaddr *) &serv_addr,
                          sizeof(struct sockaddr_in)) < 0)
    {
         logwrite(LOG_ERROR, "TELNET: ERROR on binding: %s", strerror(errno));
         exit(1);
    }
    /* Now start listening for the clients, here 
     * process will go in sleep mode and will wait 
     * for the incoming connection
     */
    if (listen(telnet->sockfd,5) < 0)
    {
         logwrite(LOG_ERROR, "TELNET: ERROR on listen: %s", strerror(errno));
         exit(1);
    }
	
    while (1) {
		/* look for free client data */
   		for(i=0; i <= telnet->max_conn ; i++) {
    		if (telnet->client_pool[i]->sockfd == -1) {
       			break;
       		}
       	}
       	
#ifdef _WIN
       	telnet->client_pool[i]->sockfd = accept(telnet->sockfd,
       							(struct sockaddr *) &telnet->client_pool[i]->clientSocket,
       							&telnet->client_pool[i]->socketlen);
#else
       	telnet->client_pool[i]->sockfd = accept(telnet->sockfd,
       							(struct sockaddr *) &telnet->client_pool[i]->clientSocket,
       							(unsigned *) &telnet->client_pool[i]->socketlen);
#endif
        if (telnet->client_pool[i]->sockfd < 0) {
            logwrite(LOG_ERROR, "TELNET: ERROR on accept: %s", strerror(errno));
            exit(-1);
        }
     
        logwrite(LOG_XTREME, "TELNET: Client: S:%d, L:%d",
        			telnet->client_pool[i]->sockfd, telnet->client_pool[i]->socketlen);
       	if (telnet->clients < telnet->max_conn) {
       		telnet->clients++;
       		
       		logwrite(LOG_XTREME, "TELNET: Clients+: %d (clientid: %d)", telnet->clients, i);
			
#ifdef HAVE_LIBPTHREAD
       		rc = pthread_create(&telnet->client_pool[i]->thread, 
       							&telnet->client_pool[i]->attr,
       							telnetClient,
       							(void *)telnet->client_pool[i]);
       		if (rc) {
       			logwrite(LOG_ERROR, "TELNET: ERROR; return code from pthread_create() is %d", rc);
       			exit(-1);
       		}
#else
	    telnetClient(telnet->client_pool[i]);
#endif
       	} else {
       		sock_write(telnet->client_pool[i]->sockfd, limitstr, strlen(limitstr));
        	close(telnet->client_pool[i]->sockfd);
#ifdef _WIN
		WSACleanup();
#endif
        	telnet->client_pool[i]->sockfd = -1;
       	}
    } /* end of client listening loop */
    
#ifdef HAVE_LIBPTHREAD
    pthread_exit(NULL);
#endif
}

void telnetCloser(telnet_t *t)
{
	int i;
	for(i=0; i <= t->max_conn ; i++ ){
       	logwrite(LOG_DEBUG, "TELNET: pool cleaned client: %d", i);

    	if (t->client_pool[i] != NULL) {
	    	if (t->client_pool[i]->sockfd != -1) {
		    close(t->client_pool[i]->sockfd);
		    t->clients++;
	    	}
    		free(t->client_pool[i]);
    	} else {
    		continue;
    	}
    	
    }
	
	logwrite(LOG_DEBUG, "TELNET: server socket close");
	close(t->sockfd);
}
